
import sqlite3, inspect
# from tesla.modal import CharField, NumberField, BooleanField, FileField

# class

DB_ADDR = 'database.db'

class DB:
    
    def __init__(self, table_name, objs) -> None:
        self.table_name = table_name
        self.db = DB_ADDR
        self.cursor = None
        self.is_connected = False
        self.__obj__ = objs
        
        self.__init_db__() 
        
    def __init_db__(self):
        self.db = sqlite3.connect(DB_ADDR)
        self.cursor = self.db.cursor()
        
        # print(self.__obj__)
        columns = []
        for name,t in self.__obj__:
            if type(t) == str:
                return
            
            q = ''
            # print(f'{t.value_type=}')
            if not inspect.isclass(type(t)):
                continue
            elif issubclass(type(t.value_type), str):
                q = f'VARCHR({t.max})'
            elif issubclass(type(t.value_type), int):
                q = 'INTEGER'
            elif issubclass(type(t.value_type), bool):
                q = 'BOOL'
                
            columns.append(f'{name} {q}')
        # print(f'{columns=}')
        if columns:
            QUERY = ', '.join(columns)
            command = f'''
            CREATE TABLE IF NOT EXISTS {self.table_name} (__id INTEGER PRIMARY KEY, {QUERY})
                                '''
            self.cursor.execute(command)
        
    
    def all(self, json = False,**kwargs):
        self.cursor.execute(f'''
        SELECT * FROM {self.table_name}
                            ''')
        rows = self.cursor.fetchall()
        self.db.commit()
        # print(rows)
        if json:
            l = []
            # print(dict())
            keys = ['id'] + list(dict(self.__obj__).keys())
            for r in rows:
                obj = {}
                for k, v in zip(keys, r):
                    obj[k] = v
                l.append(obj)
            return l    
        return rows
    
    def get(self, **kwargs):
        QUERY = ', '.join([f'"{k}" == "{v}"' for k, v in kwargs.items()])
        # print(QUERY)
        self.cursor.execute(f'''
        SELECT * FROM {self.table_name}
        WHERE ({QUERY})
                            ''')
        row = self.cursor.fetchone()
        self.db.commit()
        return row
    
    def create(self, json):
        # del json['id']
        objs = ''
        keys = []
        values = list()
        VR = []
        for k, v in json.items():
            keys.append(k)
            values.append(v)
            VR.append('?')
        k = ','.join(keys)    
        v = tuple(values)
        vr = ','.join(VR)
        command = f'''
        INSERT INTO {self.table_name} ({k})
        VALUES ({vr})
        '''
        self.cursor.execute(command, v)
        self.db.commit()
        self.db.close()


from tesla.router import Path
from tesla.response import HttpResponse, Redirect
from dataclasses import dataclass
import os


media_types = {
    'video': ['mp4'],
    'audio': ['mp3'],
    'image': ['png, jpg, jpeg'],
    
}

def get_media_type(ex):
    for t, es in media_types.items():
        if ex in es:
            return t 

@dataclass
class MediaFiles:
    path = './media/'
    request = None
    url = '/media/'



    def __post_init__(self):
        self.urls =  [
                Path('/{filename}/', self.file, name='file'),     
                Path('/{filename}/delete', self.delete, name='delete'),     
            ]


    def file(self, request):
        # print(request.params)
         
        filename = request.params['filename']
        filename, filetype = filename.split('>')
        filename = self.path + filename
        if not os.path.isfile(filename):
             
            return HttpResponse(request, f"invalidd address\n {filename}")
        with open(filename, 'rb') as file:
            return HttpResponse(request, file.read(), content_type=filetype)
    
    def delete(self, request):
        filename = request.params['filename']
        next = request.query.get('next')
        filename, filetype = filename.split('>')
        filename = self.path + filename
        if not os.path.isfile(filename):  
            os.remove(filename)
        return Redirect(request, next)    


 
mediafiles = MediaFiles()



from tesla.router.url import get_dynamic_url
from tesla.response import Render
import os
class Path:
    def __init__(self, path, func, name=None):
        self.path = path 
        self._func = func
        self.name = name
        
        
    
    def func(self, request):
        request.params = self.obj
        return self._func(request)
    def match(self, path):
        self.obj = {}
        self.obj = get_dynamic_url(path, self.path)
        # print(self.obj)
        if self.path == path:
            # print('....')
            # print(self.path, self.obj, path)
            # print('.....')
            return True
         
        elif self.obj:
            # print('....')
            # print(self.path, self.obj, path)
            # print('.....')
            # print(path, self.path)
            return True 
        return False

class Router:
    def __init__(self, routes=None):
        self.routes = list(routes) if routes else []
        self.current = None
        self.request = None


    def add_routes(self, paths):
        self.routes.append(*paths)
        # for r in self.routes:
            # print(r.path)

    def get_route(self, path):
        self.current = path
        for route in self.routes:
            # print(route.path) 
            if route.match(self.current):
                
                return route.func
        return None
    
router = Router() 


from tesla.pyhtml import PYHTML, CT
from tesla.pyhtml.tags import CSS


def ErrorPage(request, logs, title, heading1, heading2, debug):

    doc = PYHTML()
    head, body = doc.create_doc()
    # css
    style = CT('style', CSS('*', margin='0', boxSizing='border-box').css(),
               CSS('body', backgroundColor='floralwhite').css())
    h1_s = CSS('h1',
               backgroundColor='blue',
               padding='15px',
               color='white',
               borderBottom='10px solid black'

               )
    li_s = CSS('li',
               backgroundColor='lightblue',
               listStyle='none',
               padding='5px 10px',
               width='260px',
               margin='2px 0'
               )

    req_s = CSS('#req_body',
                borderTop='10px solid black',
                margin='10px 0'
                )
    style.append(h1_s.css(), li_s.css(), req_s.css())
    # head
    head.append(style, CT('title', title))

    # body
    h1 = CT('h1', title)
    p = CT('pre', CT('xmp', heading1),  style=CSS(
        margin='10px',
        marginLeft='40px',
        fontSize='x-large'
    ).css())

    ul = CT('ul')

    req_p = CT('p', 'Request body',  style=CSS(
         margin='10px',
         marginLeft='40px',
         fontSize='x-large'
         ).css())
    ul_req = CT('ul')
    req_b = CT('div', req_p, ul_req, id='req_body')
    for k, v in request.environ.items():
        x = CT('xmp', f'{k} = {v}')
        ul_req.append(CT('li', x, style=CSS(
            width='fit-content'
        ).css()))

    for l in logs:
        # print(route)
        x = CT('xmp', l)
        # if route.name:
        #     x.append(f'[name={route.name}]')
        ul.append(CT('li', x))

    if debug:
        body.append(h1, p, ul, req_b)
    else:
        body.append(title)
    return doc
    ...



from tesla.database.localdb import global_db
from .modal import User
# from http.cookiejar import Cookie


class Anonymous:
    pass

    def __str__(self):
        return 'Anonymous User'
    
    
    def get(self, key):
        return self.__str__()

class Authentication:
    def __init__(self):
        self.ANONYMOUS = Anonymous()
        self.user = None
        self.model = User
        self.session_id = ''
        
        pass
    
    def get_user(self):
        if self.user is None:
            return self.ANONYMOUS
        return self.user
    def set_user(self, obj):
        self.user = obj
        return self.user

    def authenticate(self, session_id, session):
        user_session = session.get(session_id)
        # print(f'{user_session=}')
        # print(user_session)
        self.session_id = session_id
        if not user_session:
            return False
        user = self.model.get(id=user_session.get('__id'))
        self.set_user(user)

     
        



from dataclasses import dataclass
from tesla.exceptions import DatabaseException
from tesla.modal import Model, PasswordField, EmailField
from argon2 import PasswordHasher 



class UserBaseModal(Model):
 
    def save(self):
        
        # check if no User with this username
        # exis
        
        existing_user = self.get(username = self.username)
        # print(existing_user.id != self.id)
        ph = PasswordHasher()
        if existing_user and existing_user.id != self.id :
            raise DatabaseException('User with this username already exists')
        # hashes the password
        elif existing_user == None:
            password = ph.hash(self.password)
            self.password = password
        elif ph.check_needs_rehash(self.password):
            password = ph.hash(self.password)
            self.password = password
        # elif ph.    
        return super().save()
    

class User(UserBaseModal):
    email = EmailField()
    password = PasswordField()


from pathlib import Path as Pa

from .collections import Collection

BASE_DIR = Pa(__file__).resolve().parent

abs_path = BASE_DIR

collections_manage = Collection({})

def register_collections(*model):
    for m in model:
        collections_manage.add(m)




class Paginator:
    def __init__(self, ls, limit=20, page=1, json=False) -> None:
        self.ls = ls
        self.limit = limit
        self.page = page
        self.json = json
        
        self.__categorize__()
        
    def __categorize__(self):
        self.data = []
        self.pages = 0
        temp = []
        l = 0
        for i in self.ls:
            if l == self.limit:
                self.data.append(temp)
                self.pages += 1
                temp = []
                l = 0
            if self.json:
                i = i.json()    
            temp.append(i)
            l += 1
        if len(temp) != 0:    
            self.data.append(temp) 
            self.pages += 1   
           
    
    def next(self):
        if len(self.data) <= self.page:
            return None
        p = self.page + 1
        return f'?page={p}'
        
        ...
    
    def previous(self):
        if self.page <= 1:
            return None
        p = self.page - 1
        return f'?page={p}'
        ...
        
    def current(self):
        if len(self.data) >= self.page and self.page > 0:
            return self.data[self.page-1]
        return []
        ...     
        
    def pages_lists(self):
        return [i for i in range(self.page - 2, self.page + 3) if (i > 0 and i <= self.pages) ]           



import http.server
import socketserver


def server(path):
    class MyHttpRequestHandler(http.server.SimpleHTTPRequestHandler, ):
    #def __init_
       
            def do_GET(self):
                if self.path == '/':
                   self.path = path
                return http.server.SimpleHTTPRequestHandler.do_GET(self)
               
    return MyHttpRequestHandler
        




 
from tesla import router





# 
def get_dynamic_url(url, path, o={}): 
    if (path.startswith('/static') and url.startswith('/static')) or (path.startswith('/media') and url.startswith('/media')):
        filename = url[7:]
        o['filename'] = filename
        # print('saaa') 
        return o
    # splitting the urls into a list
    sp_u = url.split('/')
    sp_p = path.split('/')
    
    # checking if the their length is not same
    if len(sp_u) != len(sp_p):
        return None     
    # print(sp_p, sp_u)    
    # getting the variables
    for i,p in enumerate(sp_p):
        if "{" and "}" in p:
            v = p[1:-1]
            o[v] = sp_u[i]
        elif p != sp_u[i]:
            return None
    return o



def Mount(ent, urls, app_name=None):
    for p in urls:
        if app_name != None:
            p.name += f'_{app_name}'
        p.path = ent + p.path
        # print(p.path)
        router.router.add_routes([p])           



from tesla.auth.modal import UserBaseModal
from tesla.modal import Model, CharField, FileField, TextField, EmailField, PasswordField,DateField, BooleanField, ImageField, NumberField, PositiveNumberField, NegativeNumberField

from dataclasses import dataclass

# class Permission(Model):
#     can_edit = BooleanField()
#     can_create = 
            

class User(UserBaseModal):
    
    username = CharField(min=4, max=10)
    email = EmailField(required=True)
    password = PasswordField(min=8, max=16, required=True)
    
    dob = DateField()
    bio = TextField()
    age = PositiveNumberField()
    dept = NegativeNumberField()
    account = NumberField()
    profile_pic = ImageField(upload_to='users_files')
    profile_doc = FileField(upload_to='users_files')
    
    is_super_user = BooleanField()
    
    
    @classmethod
    def __meta__(self):
        
        return ('id', 'username', 'email')
    
    
    def __str__(self) -> str:
        return self.username


from datetime import datetime
import random as r
import string
import os
from dataclasses import dataclass
import cgi
import cgitb
cgitb.enable()

# from tesla import TeslaApp
# from Cookie import SimpleCookie
# from time import t


@dataclass
class TemporaryFile:

    def __init__(self, file, filename, path, *args, **kwargs) -> None:
        self.file = file

        self.filename = filename
        self.path = path

        # print(self.filename, self.file)

    def save(self, upload_to=''):
        # print('file obj',self.file)
        if not upload_to.startswith('/') and upload_to != '':
            upload_to = '/' + upload_to

        path = self.path + upload_to + '/' + self.file.type

        def fbuffer(f, chunk_size=10000):
            while True:
                chunk = f.read(chunk_size)
                if not chunk:
                    break
                yield chunk
        fn = os.path.basename(self.filename)
        if not os.path.isdir(path):
            os.makedirs(path)
        if os.path.isfile(path + '/' + fn):
            name, ex = fn.split('.')
            fn = name + str(int(datetime.now().timestamp())) + '.' + ex
        with open(path + '/' + fn, 'wb+') as f:
            for chunk in fbuffer(self.file.file):
                f.write(chunk)
        return '/' + path + '/'+fn + f'>{self.file.type}'


class PostBody:
    def __init__(self, data):
        self.data = data

    def get(self, key):
        value = self.data.get(key)
        # if type(value) == str:
        #     print(dir(cgi))
        #     value = cgi.escape(value)

        return value

    def addlist(self, key, value):
        if type(self.get(key)) == str or type(self.get(key)) == int:
            fs = self.get(key)
            self.set(key, [fs])
        lst = self.get(key)
        lst.append(value)
        self.set(key, lst)

    def set(self, key, value):
        self.data[key] = value

    def __iter__(self):
        for key, value in self.data.items():
            yield key, value

    # def __dict__(self):
    #     o = {}
    #     for k,v in self:
    #         o[k] = v
    #     return o


def parse_cookies(environ):
    cookies = {}
    cookie_string = environ.get('HTTP_COOKIE', '')
    for chunk in cookie_string.split(';'):
        if '=' in chunk:
            key, value = chunk.split('=', 1)
            key, value = key.strip(), value.strip()
            cookies[key] = value
    return cookies


def parse_query_string(query_string):
    query = {}
    for param in query_string.split('&'):
        if '=' in param:
            key, value = param.split('=', 1)
            key, value = key.strip(), value.strip()
            query[key] = value
    return query


def parse_form_data(environ):
    try:
        request_body_size = int(environ.get('CONTENT_LENGTH', 0))
    except ValueError:
        request_body_size = 0
    request_body = environ['wsgi.input'].read(request_body_size)
    form_data = {}
    for param in request_body.decode().split('&'):
        if '=' in param:
            key, value = param.split('=', 1)
            key, value = key.strip(), value.strip()
            form_data[key] = value
    return form_data


def parse_file_upload(environ):
    files = {}
    try:
        request_body_size = int(environ.get('CONTENT_LENGTH', 0))
    except ValueError:
        request_body_size = 0
    request_body = environ['wsgi.input'].read(request_body_size)
    boundary = environ.get('CONTENT_TYPE', '').split('=')[-1]
    if not boundary:
        return files
    parts = request_body.split(boundary.encode())[1:-1]
    for part in parts:
        lines = part.decode().split('\r\n')
        name = None
        filename = None
        content_type = None
        data = b''
        for line in lines:
            if line.startswith('Content-Disposition:'):
                name = line.split(';')[-1].split('=')[-1][1:-1]
                filename = line.split(';')[-2].split('=')[-1][1]
            elif line.startswith('Content-Type:'):
                content_type = line.split(':')[-1].strip()
            elif line == '':
                data = b''.join(lines[lines.index(line)+1:])
        if name and data:
            files[name] = {
                'filename': filename,
                'content_type': content_type,
                'data': data,
            }
    return files


class _Request:
    def __init__(self, environ, start_response, app, authentication, context, session, auth_model):
        self.environ = environ
        self.start_response = start_response
        self.app = app
        # self.csrf_token = csrf_token
        self.authentication = authentication
        self.context = context
        self.session = session
        self.auth_model = auth_model
        self.method = environ['REQUEST_METHOD']
        self.path = environ['PATH_INFO']
        self.headers = {k[5:]: v for k,
                        v in environ.items() if k.startswith('HTTP_')}
        self.cookies = parse_cookies(environ)
        self.query_string = environ.get('QUERY_STRING', '')
        self.GET = parse_query_string(self.query_string)
        self.POST = parse_form_data(environ)
        self.FILES = parse_file_upload(environ)

    def pass_csrf(self):
        if self.method == 'POST':
            csrf_token = self.headers.get('X-CSRF-TOKEN', '')
            if not csrf_token:
                raise Exception('Missing CSRF token')
            if csrf_token != self.csrf_token:
                raise Exception('Invalid CSRF token')


class Request:
    def __init__(self, environ, start_response, app, authentication, context, session, auth_model):
        self.app = app
        # self.csrf = csrf
        self.csrf_check = False
        self.params = {}
        self.environ = environ
        self.start_response = start_response
        self.cookies = {}
        self.context = context
        self.session = session
        self.auth_model = auth_model

        self.headers = []

        self.http_host = environ.get('HTTP_HOST')
        self.http_user_agent = environ.get('HTTP_USER_AGENT')
        self.http_cookie = environ.get("HTTP_COOKIE")
        self.lang = environ.get('LANG')
        self.method = environ.get('REQUEST_METHOD')
        self.path = environ.get('PATH_INFO')
        self.host_address = environ.get('HTTP_HOST')
        self.gateway_interface = environ.get('GATEWAY_INTERFACE')
        self.server_port = environ.get('SERVER_PORT')
        self.remote_host = environ.get('REMOTE_HOST')
        self.content_type = environ.get('CONTENT_TYPE')
        self.content_length = environ.get('CONTENT_LENGTH')
        self.body = environ.get('BODY')
        self.query_string = environ.get('QUERY_STRING')
        self.server_protocol = environ.get('SERVER_PROTOCOL')
        self.server_software = environ.get('SERVER_SOFTWARE')

        self.parse_cookie()
        self.set_session_id()
        # auth request's
        # print(f'{self.session_id=}')
        authentication.authenticate(self.session_id, self.session)

        self.user = authentication.get_user()
        self.is_authenticated = self.user != authentication.ANONYMOUS
        self.session_id = authentication.session_id
        self.parse_qs()

        self.GET = parse_query_string(self.query_string)
        self.POST = parse_form_data(environ)
        self.FILES = parse_file_upload(environ)
        # self.pass_csrf()
        # self.csrf_middleware()
        # pass

    def set_session_id(self):
        # found = False
        session = self.get_cookie('user_session')
        if session:
            if ';' in session:
                session = session.replace(';', '')
            self.session_id = session
            # print(session)
            return
        session = ''.join(r.sample(string.ascii_letters, 50))
        self.session_id = session
        self.set_cookie('user_session', session,  {'max-age': 1296000})
        
    def pass_csrf(self):

        if self.method in ['POST', 'PUT', 'DELETE']:
            # print(self.csrf_check)
            if self.csrf_check == False:
                return
            
            # self.x_csrf_token = environ.get('HTTP_X_CSRF_TOKEN')
            csrf_token = self.post.data.get(
                'csrfmiddleware', None) or self.environ.get('HTTP_X_CSRF_TOKEN') or self.get_cookie('csrf_middleware_token')
            
            if csrf_token is not None:
                if csrf_token in self.app.csrf_tokens:
                    # print(self.dynamic_csrf)
                    if self.dynamic_csrf:
                        self.app.csrf_tokens.remove(csrf_token)
                        self.csrf_token = self.app.csrf_tokens[r.randint(
                            0, len(self.app.csrf_tokens)-1)]

                        self.app.csrf_tokens.append(
                            ''.join(r.sample([*string.ascii_letters,  *string.hexdigits], 55)))
                    return None
                # self.csrf = self.app.csrf_tokens[r.randint(0, len(self.app.csrf_tokens)-1)]
            # print(self.csrf)

            if not csrf_token:
                raise Exception('Missing CSRF token')
            elif csrf_token not in self.app.csrf_tokens:
                raise Exception('Invalid CSRF token')
        else:
            # elif csrf_token == None:
            # if self.dynamic_csrf:
            csrf_token = self.app.csrf_tokens[r.randint(
                0, len(self.app.csrf_tokens)-1)]
            self.csrf_token = csrf_token
            # self.csrf_token = self.app.csrf_tokens    
            self.set_header('Set-cookie', f'csrf_middleware_token={self.csrf_token}')

    def csrf_middleware(self):
        # print(self.environ.get('wsgi.csrf_token'))
        print(self.FILES, self.POST)
        # return
        csrf_token = self.cookies.get('csrf_middleware')
        if self.method == 'POST':
            print(csrf_token)
            if not csrf_token:
                raise Exception('Missing CSRF token')
            elif csrf_token not in self.app.csrf_tokens:
                raise Exception('Invalid CSRF token')
            # self.csrf_token = csrf_token
            # else:
            #     csrf_token = self.app.csrf_tokens[r.randint(0, len(self.app.csrf_tokens)-1)]
            #     self.csrf_token = csrf_token
            #     self.set_header('Set-Cookie', f'csrf_middleware={csrf_token};max-age=300;path=/')
        elif csrf_token == None:
            csrf_token = self.app.csrf_tokens[r.randint(
                0, len(self.app.csrf_tokens)-1)]
            self.csrf_token = csrf_token
            self.set_header('Set-Session', f'csrf_middleware={csrf_token}')

    def parse_qs(self):
        # print(self.environ)
        qs = self.query_string.split('&')
        self.query = {}
        for q in qs:
            v = ''
            k = q
            if '=' in q:
                k, v = q.split('=')
            self.query[k] = v.replace('+', ' ')

        if self.method not in ['POST', 'PUT']:

            return

        self.post = PostBody({})
        environ = self.environ
        field_storage = cgi.FieldStorage(
            fp=environ['wsgi.input'],
            environ=environ,
            keep_blank_values=True
        )

        for item in field_storage.list:
            if not item.filename:
                if type(item.value) not in [str, int, float]:
                    continue

                if self.post.get(item.name):
                    self.post.addlist(item.name, item.value)
                    continue
                self.post.set(item.name, item.value)
            else:
                # print((item.type))
                self.post.set(item.name, TemporaryFile(
                    file=item, filename=item.filename, path=self.app.media_file))

    def clear_cookie(self):
        self.cookie = []

    def set_cookie(self, k, v, kwargs={}):
        default = {
            'path': '/',
            'max-age': 1296000,
        }
        kwargs = {**default, **kwargs}
        # print(kwargs)
        p = ';'.join([f'{k}={v}' for k, v in kwargs.items()])
        self.cookies[k] = v
        # print(p)
        self.set_header('Set-Cookie', f'{k}={v};{p}')

    def parse_cookie(self):
        if self.http_cookie == None:
            return
        cookies = self.http_cookie.split(' ')
        for cookie in cookies:
            k, v = cookie.split('=')
            self.set_cookie(k, v, {})

    def set_header(self, key, value):
        self.headers += [(key, value)]

    def get_cookie(self, key):
        return self.cookies.get(key)



from tesla.pyhtml.tags import CT, get_children, get_tag_style_attr
from tesla.pyhtml.watcher import Watcher
from tesla.pyhtml.server import server


import socketserver
import http.server

import os
import threading
class PYHTML:
    def __init__(self,html = True,css = False,js = False, path='./'
    ):
        self.html = html
        self.css = css
        self.js = js     
        self._path = path
        self.tag = CT("html")
        if not os.path.isdir(path):
            os.mkdir(path)
        
        
    def generate_files(self):
        if self.css:
            elems = self.tag.args
            
            inline_css = []
            
            for elem in elems:
               for child in get_children(elem):
                   if child.tag == "style":
                       inline_css.extend(child.args[:])
                       child.removeSelf("link", rel="stylesheet", href=f"./style.css")
               inline_css.extend(get_tag_style_attr(elem))
            with open(f"{self._path}/style.css", "w") as file:
                for css in inline_css:
                    
                    file.write(str(css))

        if self.html:
            with open(f'{self._path}/index.html', 'w') as file:
                file.write(format(self.tag.html()))
                
    def __str__(self):
        return format(self.tag.html())
    def allElems(self):
        elems = []
        head = self.tag
        for elem in head.args:
            elems.extend(get_children(elem))
        return elems

    def create_doc(self):
       
        body = CT("body")
        head = CT("head")
        self.tag.append(head, body)
        
        return head, body
        
    def preview(self):
        #self.generate_files()
        path = self._path
        print(f"{'':.^40}")
        print(f":{'PyHtml preview':^38}:")
        print(f"{'':.^40}")
       
        
        #watcher = Watcher("./")
        
        handler_object = server(path)
        PORT = 1234
        with socketserver.TCPServer(("", PORT),handler_object) as my_server:
            socketserver.TCPServer.allow_reuse_address = True
            print(f"Server started at http://localhost:{PORT}")
        # Star the server
            t1 = threading.Thread(target=Watcher, args=["./",self.generate_files, my_server])
            t1.start()
        #my_server.allow_reuse_address = True
            my_server.serve_forever()
    




from tesla.router import Path
from tesla.response import Render, JsonResponse,Response

from tesla import TeslaApp
import os
import sys
import ast


def join_path(paths):
    count = 0
    r = ''
    for p in paths:
        r += p


def wrapper(pg, backends=[], context={}):
    # TeslaApp.backend
    pth = os.path.join(TeslaApp.backend)
    sys.path.append(pth)
    import views

    def req(request):
        b = {}
        redirect = False
        for v in backends:
            func = getattr(views, v)
            obj = func(request)
            if type(obj) != dict:
                if issubclass(type(obj), Response):
                    redirect = obj
                    break
            b = {**context, **b, **obj}
        # print(redirect)
        if redirect:
            
            return redirect
        return Render(request, pg, b)
    return req


def get_pages(address) -> list:
    routes = []
    pgs = os.listdir(address)

    for pg in pgs:
        if os.path.isdir(address / pg):
            pass
        else:
            name, *views_file, _ = pg.split('.')
            context = {}

            if name.startswith('index'):
                name = name.replace('index', '')
            if '_' in name:
                n = ''
                for p in name.split('_'):
                    if '[' and ']' in p:
                        n = os.path.join(n, '{' + p[1:-1] + '}')
                    else:
                        n = os.path.join(n, p)
                name = n.replace('\\', '/')
            p = Path(name, wrapper(pg, context=context,
                     backends=views_file), name=name)
            routes.append(p)

    return routes


def config_api_route(route):
    pth = os.path.join(TeslaApp.base_dir, 'api')
    sys.path.append(pth)
    try:
        import api
        apis = [a for a in dir(api) if not a.startswith('__')]

        def wrapper(api):
            def req(request):
                b = api(request)
                return JsonResponse(request, b)
            return req
        routes = [Path(a, wrapper(getattr(api, a)), name=a) for a in apis]
        TeslaApp.mount('/'+route+'/', routes, route)
    except Exception as e:
        raise Exception(e)



from tesla.context import global_context
from tesla.media import mediafiles
from tesla.request import Request
from tesla.response import Http404Response, Http500Response,Render
from tesla.router import router
from tesla.auth import Authentication
from tesla.session import Session
from tesla.static import staticfiles
from tesla.middleware import middlewares
from tesla.auth.modal import User

import string
import random as r



def csrf():
    ls = []
    for i in range(10):
        s = ''.join(r.sample([*string.ascii_letters,  *string.hexdigits],55))
        ls.append(s)
    return ls
        

import random
import string

class _App:
    def __init__(self):
        self.debug = True
        self.dynamic_csrf = True
        self.router = router
        self.context = global_context
        self.templates_folders = []
        self.registered_models = []
        self.auth_model = User
        self.csrf_tokens = csrf()
        self.middlewares = middlewares
        self.media_file = 'media'
        self.base_dir = None
        self.db = 'tesla.database.SQL'
        

        self.mount('/static', staticfiles.urls, app_name='static')
        self.mount('/media', mediafiles.urls, app_name='static')

    def set_auth_model(self, model):
        self.auth_model = model
        
    def set_routes(self, routes: list):
        for path in routes:
            self.router.add_route(path)

    def mount(self, path, urls, app_name):
        for p in urls:
            if app_name != None:
                if p.name:
                    p.name += f'_{app_name}'
                p.path = path + p.path
            self.router.add_routes([p])           

    def __call__(self, environ, start_response):

        self.authentication = Authentication()
        self.session = Session({})
        if self.auth_model is not None:
            self.authentication.model = self.auth_model
        
        request = Request(environ, start_response, self, self.authentication, self.context, self.session, self.auth_model)
        request.dynamic_csrf = self.dynamic_csrf
        staticfiles.request = request
        
        try:
            for mid in self.middlewares.middlewares():
                # print(request.csrf_check)
                err = mid(request) 
                if err:
                    return Http500Response(request, err, self.debug).make_response()
            request.pass_csrf()
            # print(request.csrf_check)
            
            func = self.router.get_route(request.path)
            if func is not None:
                # print(func)
                response = func(request)
                return response.make_response()


            else:
                return Http404Response(request, self.debug, self.router.routes).make_response()
        except Exception:
            import traceback
            print(traceback.format_exc())
            return Http500Response(request, traceback.format_exc(), self.debug).make_response()





class __App: 
    def __init__(self):
        self.debug = True
        self.router = router
        self.context = global_context
        self.templates_folders = []
        self.registered_models = []
       
        self.auth_model = UserBaseModal
        self.csrf_tokens = csrf()
        self.middlewares = middlewares
        self.media_file = 'media'

        # self.mount('/static', staticfiles.urls, app_name='static')
        # self.mount('/media', mediafiles.urls, app_name='static')
        
        pass
    
    def set_auth_model(self, model):
        self.auth_model = model
    def set_routes(self, routes: list):
        for path in routes:
            self.router.add_route(path)

    def mount(self, path, urls, app_name):
        # print(path)
        for p in urls:
            # print(p)
            if app_name != None:
                if p.name:
                    p.name += f'_{app_name}'
                p.path = path + p.path
            self.router.add_routes([p])           

    def __call__(self, environ, start_response):
        
        self.authentication = Authentication()
        self.session = Session({})
        if self.auth_model is not None:
            self.authentication.model = self.auth_model
        
        request = Request(environ, start_response,self, self.csrf_tokens[r.randint(0, len(self.csrf_tokens)-1)], self.authentication, self.context, self.session, self.auth_model)
        staticfiles.request = request
        
        try:
            # request.pass_csrf()
            for mid in self.middlewares.middlewares():
                err = mid(request) 
                if err:
                    return Http500Response(request, err, self.debug).make_response()
            
            func = self.router.get_route(request.path)
            if func is not None:
                
                response = func(request)
                # response.templates_folders.extend(self.template_folders)
                # print(type(response))
                return response.make_response()
 
            else:
                return Http404Response(request, self.debug, self.router.routes).make_response()
        except Exception:
            import traceback
            # print('jjujujuju')
            # raise Exception(e)
            print(traceback.format_exc())
            return Http500Response(request, traceback.format_exc(), self.debug).make_response()
    







TeslaApp = _App()    
    
    


from dataclasses import dataclass, field

@dataclass
class Collection:
    models : dict = field(default_factory=dict)
    
    
    def add(self, model):
         if model.__name__ in self.models:
             return self.models[model.__name__]
         self.models[model.__name__] = {
             'name': model.__name__,
             'counts': model.size()[0],
             'size': model.size()[1]
         }
         return model
     
    def remove(self, model):
        if model.__name__ in self.models:
            del self.models[model.__name__] 
            return None
        return None
    
    def colls(self):
        return [m for m in self.models.values()]



from tesla.router import Path
from . import views

# your urls path should be here


patterns = [
    Path('', views.index, name='index'),
    Path('account/', views.admin_account, name='admin_account'),
    Path('login/', views.login, name='login'),
    Path('logout/', views.logout, name='logout'),
    Path('register/', views.register, name='register'),
    Path('reset-password/', views.reset_password, name='reset-password'),
    Path('collections', views.collections, name='collections'),
    Path('collections/{collection}/', views.collection, name='collection'),
    Path('collections/{collection}/new/', views.collection_new, name='collection_new'),
    Path('collections/{collection}/delete/', views.collection_del_all, name='collection_del_all'),
    Path('collections/{collection}/json/', views.collection_download, name='collection_download'),
    Path('collections/{collection}/{lookup}/', views.collection_obj, name='collection_obj'),
    Path('collections/{collection}/{lookup}/delete/', views.collection_del, name='collection_del')
 
]




def settype(type, default = None, onerror = ""):
    return {"type": type, "default": default, 'onerror': onerror}
    
    
User =  {
    "age" : settype(int, 0),
    "name" : settype(str, "default"),
    "hobbies" : settype(list,[2,2,3]),
    "is_admin" : settype(bool, True),
}




from dataclasses import dataclass
from tesla.functions import media_file, static_url, url



@dataclass
class Context:
    objs : dict

    def get_objs(self):
        return self.objs

    def add(self,key, value):
        self.objs[key] = value


global_context = Context({})
global_context.add('static_url', static_url)
# global_context.add('messages', messages_broker)
global_context.add('url', url)
global_context.add('media_file', media_file)


from tesla.response import Render
from tesla.pagination import Paginator

class View:
    model = None
    lookup = None
    template = None
    pagination = False
    pagination_count = 20
    response = Render
    

    def as_view(cls, **kwargs):
        
        def wrapper(request, *args, **kwargs):
            method = request.method
            if hasattr(cls, method.lower()):
                func = getattr(cls, method.lower())
                return func(request, *args, **kwargs)
            else:
                raise Exception(f'{method} Method is not allowed or provided')
            
        return wrapper
    

class DetailView(View):
    
    def get(self, request, *args, **kwargs):
        
        lk = request.params.get(self.lookup)
        all = self.model.all(models=True)
        obj = None
        for o in all:
            if getattr(o, self.lookup) == lk:
                obj = o
                break
        if issubclass(self.response, Render):
            context = {}
            context['obj'] = obj
            return self.response(request, self.template, context)
        if obj == None:
            json_obj = {}
        else:
            json_obj = obj.json()    
        return self.response(request, json_obj)

class RetrieveAllView(View):
    
    def get(self, request, *args, **kwargs):
        all = self.model.all(models=True)

        if issubclass(self.response, Render):
            context = {}
            context['objs'] = all
            return self.response(request, self.template, context)
        json_obj = []
        for o in all:
            json_obj.append(o.json())
        if self.pagination:
            page = request.query.get('page')
            if not page:
                page = 1
            paginator = Paginator(all,json=True, page=int(page), limit=1)
            json_obj = {
                'page': 1,
                'next': paginator.next(),
                'previous': paginator.previous(),
                'data': paginator.current()
            } 
        
   
        return self.response(request, json_obj)  


# from __future__ import annotations
import uuid
from copy import deepcopy
from datetime import datetime

# from tempfile import TemporaryFile
from tesla.database.jsondb import DB as JsonDB
from tesla.database.SQL import DB as SQLDB
from tesla.functions import truncate, url
from tesla.pyhtml.tags import CSS, CT
from tesla.request import TemporaryFile
from tesla.signal import signal


def property_obj(arr, d=None):
    a = []
    for l, v in arr.__dict__.items():
        if not l.startswith('__'):
            if l != 'id' and l != 'timestamp':
                if d:
                    l = (l, v)
                a.append(l)

    return a


def to_dict(arr):
    a = {}
    for k, v in arr:
        a[k] = v
    return a


def ModalId():
    cache = {}

    def generate(modal):

        if modal in cache.keys():

            cache[modal] += 1
            return cache[modal]
        cache[modal] = 1
        return 1
    return generate



class Field:

    def __init__(self, required=False, label=None) -> None:
        self.input_type = 'text'
        self.type = 'input'
        self.required = required
        self.css = CSS()
        self.name = ''
        self.params = ''
        self.default = ''
        self.label = None
        self.value_type = str

        pass

    def __pre_show__(self):
        pass

    def input(self, name, **kwargs):
        k = ' '.join([f'{a}={b}' for a, b in kwargs.items()])
        self.params += ' ' + k
        self.name = name
        if self.required:
            self.params += ' required'
        if self.type == 'textarea':
            self.css.kwargs['height'] = '200px'      # s
            self.css.kwargs['resize'] = 'none'      # s
        return self

    def html(self, **kwargs) -> str:
        self.input(self.name)

        self.tag = CT(self.type, **kwargs, style=self.css.css(
        ), type=self.input_type, name=self.name, value=self.default, params=self.params)
        if self.type != 'input':
            self.tag.append(self.default)
            del self.tag.kwargs['value']
        if self.label == None:
            self.label = self.name

        self.__pre_show__()
        return self.tag.html()

    # def __str__(self):
    #     self.html()
    #     return self.tag.html()

    def __repr__(self):
        return self.__class__.__name__

    def validate(self, value):
        try:
            v = self.value_type(value)
            if isinstance(v, self.value_type):
                return True
            return False
        except Exception:
            raise Exception(f'{value} is not valid type for {self.__repr__()}')


class CharField(Field):

    def __init__(self, default='', min=0, max=120, *args, **kwargs) -> None:
        self.default = default
        self.min = min
        self.max = max
        super().__init__(*args, **kwargs)

    def input(self, *args):
        kwargs = {}
        kwargs['minlength'] = self.min
        kwargs['maxlength'] = self.max
        return super().input(*args, **kwargs)


class PasswordField(CharField):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.input_type = 'password'


class ListField(Field, list):

    def __init__(self, default=[], *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.default = default
        self.type = 'select'


class TextField(Field):

    def __init__(self, default='', *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.default = default
        self.type = 'textarea'


class EmailField(CharField):

    def __init__(self, default='', min=0, max=120, *args, **kwargs) -> None:
        super().__init__(default, min, max, *args, **kwargs)
        self.default = default
        self.input_type = 'email'


class DateField(CharField):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.input_type = 'date'
        self.default = str(datetime.now())


class DictField(Field, dict):

    def __init__(self, default={}, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.default = default


class ForeignKey(Field):

    def __init__(self, related_model, required=False, label=None) -> None:
        super().__init__(required, label)
        self.default = None
        self.related_model = related_model
        self.related_model_id = None
        self.type = 'select'

    @property
    def option_models(self):
        return self.related_model.all()

    def __pre_show__(self):
        for obj in self.option_models:

            opt = CT('option', str(obj), value=obj.id)
            if self.related_model_id == obj.id:

                opt.params += ' selected'
            self.tag.append(opt)
            ...

    def input(self, name, **kwargs):
        name += '__id'

        return super().input(name, **kwargs)


class ManyToManyField(Field):

    def __init__(self, related_model, required=False, label=None) -> None:
        super().__init__(required, label)
        self.default = []
        self.related_model = related_model
        # self.option_models = related_model.all()
        self.related_model_ids = []
        self.type = 'select'

    @property
    def option_models(self):
        return self.related_model.all()

    def __pre_show__(self):
        self.tag.params += 'multiple'
        for obj in self.option_models:

            opt = CT('option', str(obj), value=obj.id)
            if obj.id in self.related_model_ids:

                opt.params += ' selected'
            self.tag.append(opt)
        return super().__pre_show__()

    def input(self, name, **kwargs):
        name += '__ids'

        return super().input(name, **kwargs)

    def related_models(self):
        return [self.related_model.get(id=id) for id in self.related_model_ids]

class BooleanField(Field):
    
    def __init__(self, required=False, label=None) -> None:
        super().__init__(required, label)
        self.input_type = 'checkbox'
        self.default = False
        self.value_type = bool
    
    def html(self, **kwargs) -> str:
        self.input(self.name)
        if self.default:
            self.params += ' checked'
        self.tag = CT('input', type='checkbox', name=self.name, params=self.params)
        # return super().html(**kwargs)
        self.__pre_show__()
        return self.tag.html()

class NumberField(Field):
    
    def __init__(self, required=False, label=None) -> None:
        
        super().__init__(required, label)
        self.input_type = 'number'
        self.default = 0 
        self.min = -9999999999
        self.max = 99999999999
        self.value_type = int
    
        
    def input(self, *args):
        kwargs = {}
        kwargs['min'] = self.min
        kwargs['max'] = self.max
        return super().input(*args, **kwargs)    

class PositiveNumberField(NumberField):
    
    def __init__(self, required=False, label=None) -> None:
        super().__init__(required, label)
        self.default = 0 
        self.min = 0
        self.max = 9999999999
    

class NegativeNumberField(NumberField):
    
    def __init__(self, required=False, label=None) -> None:
        super().__init__(required, label)
        self.default = -1 
        self.min = -999999999
        self.max = 0
    
        

class FileField(Field):
    
    def __init__(self,upload_to = '' , required=False, label=None) -> None:
        super().__init__(required, label)
        self.upload_to = upload_to
        self.accept = '*'
    
    def html(self, **kwargs) -> str:
        self.input(self.name)
        self.tag = CT('input', type='file', accept=self.accept, name=self.name, params=self.params)
        if self.default:
            p = CT('p', 'current file : ')
            old_file = CT('a',  truncate(self.default), href=self.default)
            delete_old_file = CT('a',  'remove', href=self.default + '?next={{request.path}}')
            
            p.append(old_file)
            p.append(' | ')
            p.append(delete_old_file)
            self.tag = CT('span', self.tag, p)
        # return super().html(**kwargs)
        self.__pre_show__()
        return self.tag.html()    
    def validate_file(self, file, name):
        if self.accept == '*':
            return
        file_type = self.accept[:-2]
        if file.type.split('/')[0] != file_type:
            raise Exception(f'{file.type} is not accepted for {name}') 
class ImageField(FileField):
    
    def __init__(self, upload_to='', required=False, label=None) -> None:
        super().__init__(upload_to, required, label)
        self.accept = 'image/*'

class AudioField(FileField):
    
    def __init__(self, upload_to='', required=False, label=None) -> None:
        super().__init__(upload_to, required, label)
        self.accept = 'audio/*'

class VideoField(FileField):
    
    def __init__(self, upload_to='', required=False, label=None) -> None:
        super().__init__(upload_to, required, label)
        self.accept = 'video/*'

class Model:

    id = CharField()
    timestamp = CharField()
    __db_type__ = 'sql'

    def __init__(self) -> None:
        self.assign_foreign()

        # pass

    @classmethod
    def get_collection(cls, model, dept):

        return JsonDB(collection=cls.__name__ + "/") .get_collection(model)

    @classmethod
    def create(cls, *args, **kwargs):
        # kwargs['id'] = str(uuid.uuid4())
        # kwargs['timestamp'] = str(datetime.now())
        c = cls()
        for k, v in kwargs.items():

            setattr(c, k, v)
        c = c.save()
        return c

    @classmethod
    def filter(cls, **kwargs):

        db = JsonDB(collection=cls.__name__ + "/")
        return db.filter(kwargs)

    @classmethod
    def all(cls, models=True):
        if cls.__db_type__.lower() == 'sql': 
            db = SQLDB(cls.__name__,  property_obj(cls, True))
        elif cls.__db_type__.lower() == 'json':
            db = JsonDB(collection=cls.__name__ + "/")
            
        if not models:
            return list(db.all(json=True))
        l = []
        for m in db.all():
            c = cls()
            if type(m) != dict:
                dj = {}
                k = cls().property_cls()
                k = ["i"] + k
                for k, v in zip(k, m):
                    # if i == 0:
                    #     dj['id'] = m[i]
                    dj[k] = v
                m = dj
            for k, v in m.items():

                if '__' in k:
                    mn = k.split('__')[0]
                    tt = getattr(c, mn)
                    if issubclass(type(tt), ForeignKey):
                        tt.related_model_id = v
                        setattr(c, mn, tt.related_model.get(id=v))
                        
                    if issubclass(type(tt), ManyToManyField):
                        tt.related_model_ids = []
                        for i in v:
                            tt.related_model_ids.append(i)
                        setattr(c, mn, tt.related_models())
                        # print(v)
                        # setattr(c, mn, tt.related_model.get(id=v))

                setattr(c, k, v)
            # c.assign_foreign()
            l.append(c)
        return l

    @classmethod
    def size(cls):
        if cls.__db_type__.lower() == 'sql': 
            db = SQLDB(cls.__name__,  property_obj(cls, True))
        elif cls.__db_type__.lower() == 'json':
            db = JsonDB(collection=cls.__name__ + "/")
            
        return db.size()

    @classmethod
    def get(cls, *args, **kwargs):

        if cls.__db_type__.lower() == 'sql': 
            db = SQLDB(cls.__name__, property_obj(cls, True))
        elif cls.__db_type__.lower() == 'json':
            db = JsonDB(collection=cls.__name__ + "/")
            
        js = db.get(**kwargs)
        # print(js)
        c = None
        if js != None:
            c = cls()
            for k, v in js.items():
                if '__' in k:
                    mn = k.split('__')[0]
                    tt = getattr(c, mn)
                    if issubclass(type(tt), ForeignKey):
                        tt.related_model_id = v
                        setattr(c, mn, tt.related_model.get(id=v))
                        # print(k, v)
                    if issubclass(type(tt), ManyToManyField):
                        tt.related_model_ids = []
                        for i in v:
                            tt.related_model_ids.append(i)
                        setattr(c, mn, tt.related_models())
                setattr(c, k, v)
        return c

    @classmethod
    def __subclasses(cls):
        l = cls.__subclasses__()
        r = []
        for i in l:
            r.append(i.__name__.capitalize())
        return r

    @property
    def db(self):
        if self.__db_type__.lower() == 'sql': 
            db = SQLDB(self.__class__.__name__,  property_obj(self, True))
        elif self.__db_type__.lower() == 'json':
            db = JsonDB(collection=self.__class__.__name__ + "/")
        return db

    def modal_name(self):
        return (self.__class__.__name__)

    def save(self):
        created = not bool(self.get(id = self.id))
        signal.send(self.__class__, self, created, 'pre-save')
        cls_copy = self.__class__

        j = {**self.json()}
        cls_props = to_dict(property_obj(cls_copy, True))
        props = {**cls_props, **j}


        for k, v in props.items():
            
            if '__' in k:
                ref_k = k.split('__')[0]
                if ref_k in j:
                    del j[ref_k]
            if type(v) == str and '__ids' in k:
                j[k] = [v]
            if issubclass(type(v), (ForeignKey, ManyToManyField)) or issubclass(type(cls_props.get(k)), (ForeignKey, ManyToManyField)):
                is_m2m = issubclass(type(v), ManyToManyField) or issubclass(type(cls_props.get(k)), ManyToManyField)
                # is_1t1 = issubclass(type(v), ForeignKey)
                
                kk = k + '__id'
                if is_m2m:
                    kk += 's'
                if k in j.keys():
                        del j[k]
                if issubclass(type(v), Model):
                    j[kk] = v.id
                else:
                    j[kk] = v.default
                # print(j, )
                continue
            elif issubclass(type(v), (BooleanField, DateField)):
                j[k] = v.default
                
            elif issubclass(type(cls_props.get(k)), BooleanField):
                if v == 'on':
                    j[k] = True  
                else:
                    j[k] = False        
            elif issubclass(type(v), TemporaryFile):
                _model_field = cls_props.get(k)
                _model_field.validate_file(v.file, k)
                j[k] = v.save(_model_field.upload_to)
            elif type(v) == str and v.isnumeric():
                j[k] = int(v)    
            setattr(self, k, v)

        # if not bool(self.id):
        #     self.id = str(uuid.uuid4())
        #     j['id'] = self.id
        #     self.timestamp = str(datetime.now())
        #     j['timestamp'] = self.timestamp

        # self.db.create_column(model=j, table_name=str(self.id))
        # db = JsonDB(collection=cls.__name__ + "/")
        # db = 
        self.db.create(j)

        # self.db.g
        signal.send(self.__class__, self, created,'post-save')
        
        return self

    def update(self, **kwargs):

        for k, v in kwargs.items():
            if '__' in k:
                kk = k.split('__')[0]
                tt = getattr(self, kk)
                # old_ref = ge

                if issubclass(type(tt), Model):
                    new_ref = tt.__class__.get(id=v)
                    setattr(self, kk, new_ref)
                if issubclass(type(tt), ForeignKey):

                    tt.related_model_id = v
                if issubclass(type(tt), ManyToManyField):
                    tt.related_model_ids = []
                    if type(v) == str:
                        v = [v]
                    # print(v)
                    for i in v:
                        tt.related_model_ids.append(i)

            setattr(self, k, v)
        return self

    def json(self):
        # print(self.__annotations__)
        return self.__dict__

    def delete(self):
        self.db.delete(self.id)
        return None

    def assign_foreign(self):

        cls_copy = self.__class__

        j = {**self.json()}
        cls_props = to_dict(property_obj(cls_copy, True))
        
        # for k, v in j.items():
        #     print(k, v)
        #     if '__' in k:
        #         rel_id = cls_copy.get(k).related_model_id
        #         print(rel_id)
            # setattr(self, cl, c)

    @classmethod
    def __meta__(cls):
        return ('id',)

    @classmethod
    def property_cls(cls):
        ls = property_obj(cls)

        return ls

    def admin_dis(self):
        tp = self.__meta__()
        cls_dis = self.get(id=self.id)
        f = []
        for t in tp:

            tt = getattr(cls_dis, t)

            if issubclass(type(tt), list):
                tt = ', '.join([str(m) for m in tt])
            f.append(tt)

        return f

    def __str__(self) -> str:
        return f'{self.__class__.__name__}({self.id})'




from tesla.response import JsonResponse
from tesla.pagination import Paginator


class Serializer:
    
    model = None
    
    dept = 1
    
    
    # def __init__(self) -> None:
    #     self.models = self.model.all()
    
    
    def serialize(self):
        return self.model.all(models=False)
        ...
    
class APIView:
    model = None
    lookup = None
    pagination = False
    pagination_count = 20
    response = JsonResponse
    

    def as_view(cls, **kwargs):
        
        def wrapper(request, *args, **kwargs):
            method = request.method
            if hasattr(cls, method.lower()):
                func = getattr(cls, method.lower())
                return func(request, *args, **kwargs)
            else:
                raise Exception(f'{method} Method is not allowed or provided')
            
        return wrapper
    

class RetrieveAPIView(APIView):

    def get(self, request, *args, **kwargs):
        all = self.model.all(models=True)

        
        json_obj = []
        for o in all:
            json_obj.append(o.json())
        if self.pagination:
            page = request.query.get('page')
            if not page:
                page = 1
            paginator = Paginator(all ,json=True, page=int(page), limit=self.pagination_count)
            data = paginator.current()
            json_obj = {
                'count': len(data),
                'all': len(all),
                'next': paginator.next(),
                'previous': paginator.previous(),
                'data': data
            } 
        
   
        return self.response(request, json_obj)  
    
class CreateRetrieveAPIView(RetrieveAPIView):
    
    def post(self, request, *args, **kwargs):
        
        self.model.create(**request.post.data)
        
        return self.get(request, *args, **kwargs) 
    

class UpdateDelete(APIView):
    
    def get(self, request, *args, **kwargs):
        lookup = request.params.get(self.lookup)
        all = self.model.all()
        model = None
        # print(lookup)
        for i in all:
            js = i.json()
            if js[self.lookup] == lookup:
                # print(js.get(self.lookup) == lookup)
                model = i.json()
                break
        return self.response(request, model) 
    
    def put(self, request, *args, **kwargs):
        lookup = request.params.get(self.lookup)
        all = self.model.all()
        model = None
        # print(lookup)
        for i in all:
            js = i.json()
            if js[self.lookup] == lookup:
                # print(js.get(self.lookup) == lookup)
                i.update(**request.post.data)
                i.save()
                model = i.json()
                break
        return self.response(request, model)
    
    def delete(self, request, *args, **kwargs):
        lookup = request.params.get(self.lookup)
        all = self.model.all()
        model = None
        # print(lookup)
        for i in all:
            js = i.json()
            if js[self.lookup] == lookup:
                # print(js.get(self.lookup) == lookup)
                i.delete()
                break
        return self.response(request, None)  




from tesla.functions import redirect
from tesla.messages import messages_broker


def login_required(path = 'login'):
    # print(args, kwargs)
    def decorator(func):
        def inner(request, *args, **kwargs):
            if request.is_authenticated:
                return func(request, *args, **kwargs)
            messages_broker.add_message(request,'You must login')
            return redirect(request, f'{path}')
        
        return inner
    return decorator


from uuid import uuid4
from tesla.pyhtml.formater import format
from tesla.pyhtml.js import document

import os


def get_children(elem, rec=False):
    children = []
    for child in elem.args:
        children.append(child)
        
        if rec:
            if type(child) != str and len(child.args) != 0:
                children.extend(get_children(child, rec))
    return children


def get_tag_style_attr(child):
     inline_css = []
     if child.kwargs.get("style"):
              styles = "[pyid='" + child.kwargs.get("pyid") + "']{" + child.kwargs.get('style') + "}"
              #print(child.kwargs.get("style")
            
              del (child.kwargs["style"])
              #delattr(child.kwargs, 'style')
              #print(child.kawrgs)
              inline_css.append(styles)
     if len(child.args) > 0:
        for c in child.args:
            if type(c) != str:
                inline_css.extend(get_tag_style_attr(c))
     return inline_css

class CSS:
    def __init__(self, target= None, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.target =target
        
    def css(self):
        return self.interprete()
    def interprete(self):
        css = ""
        for p, a in self.kwargs.items():
            property = self.tran(p)
            attribute = self.tran(a)
            css += f"{property}: {attribute};"
        if self.target != None:
            css = self.target + "{" + css + "}"
        return css
    def tran(self, txt):
        r = ""
        for i,l in enumerate(txt):
            if l.isupper():
                if i != 0:
                    l = f'-{l}'
                l = l.lower()
            r += l
        return r

class CSSGroup:
    def __init__(self, name, *args, **kwargs):
        self.name = name
        self.args = args
        self.kwargs = kwargs
        self.assignClass()
    def assignClass(self):
        for arg in self.args:
            arg.kwargs["class"] = self.name
    def __str__(self):
        return CSS(f".{self.name}", **self.kwargs).css()
    def __repr__(self):
        return self.__str__()

 
     
     
class CT:
    def __init__(self, tag, *args, params='', **kwargs ):
        self.tag = tag
        self.args = args
        self.kwargs = kwargs
        self.params = params
        # self.kwargs["pyid"] = "pyhtml_" + str(uuid4()).split("-")[0]
        # self.doc = document.get('[pyid="'+self.kwargs.get('pyid')+'"]')
    def __repr__(self):
        return f"<{self.tag} children={get_children(self, False)} />" 
    def html(self):
        # pr
        r = " ".join(map(lambda elem: str(elem), self.args))
        # print(self.args)
        kwargs = ''
        for k, v in self.kwargs.items():
            kwargs += f" {k}='{v}'"
        # print(self.params)   
        
        return f"<{self.tag}{kwargs}{self.params}>{r}</{self.tag}>"


    def __str__(self):
        return self.html()    
    def append(self, *tags):
        self.args = list(self.args)
        for tag in tags:
            self.args.append(tag)
        return None
    def removeSelf(self,tag =None, *args, **kwargs):
        self.args = []
        self.kwargs = kwargs
        if tag:
            self.tag = tag
        return None

    def onClick(self, event):
        return self.doc.onClick(event)
        
    def value(self):
        return self.doc.getValue()

    def setattr(self, key, value):
        self.kwargs[key] = value
        return None

class SCT(CT):
    def __init__(self, tag,  *args, **kwargs):
        super().__init__(tag,  *args, **kwargs)
    def html(self):
        r = " ".join(self.args)
        kwargs = ''
        for k, v in self.kwargs.items():
            kwargs += f" {k}='{v}'"
        return f"<{self.tag}{kwargs} />"
 
     
    

        
        
class TextNode:
  def __init__(self, text):
      self.text = text
  def html(self):
      return self.text
  


# from tesla import TeslaApp
import tesla
from tesla.pages import ErrorPage
from tesla.request import Request
from tesla.messages import messages_broker


import os

from jinja2 import Environment, FileSystemLoader, PackageLoader, select_autoescape, ChoiceLoader


class Response:
    def __init__(self, request: Request, status_code: str, content_type: str):

        self.status_code = status_code
        self.start_response = request.start_response
        self.content_type = content_type
        # print(request.get_cookie())
        self.headers = [('Content-type', self.content_type)]
        self.headers += request.headers

        self.response_content = []
        self.templates_folders = ['./']
        self.templates_folders.extend(tesla.TeslaApp.templates_folders)

    def make_response(self):
        self.start_response(self.status_code, self.headers)
        return self.response_content

    def parse_cookie(self, cookie):
        cookie = cookie.split(';')
        obj = {}
        for c in cookie:
            c = c.split('=')
            obj[c[0]] = c[1]
        cookie = obj
        return obj.items()


class Render(Response):

    def __init__(self, request: Request, content, context={},  status_code='200 OK', content_type='text/html'):

        # request.cookie = self.parse_cookie(request.http_cookie)
        request.headers += [('cache-control', 'private, no-cache')]
        super().__init__(request, status_code, content_type)
        # env = Environment(loader=PackageLoader('jinja2', '/'))

        # if self.is_available(content) is not None:
        # template = env.get_template(self.is_available(content))
        file_loader = FileSystemLoader(self.templates_folders)
        loader = ChoiceLoader([
            # PackageLoader("tesla.admin", 'templates'),
            FileSystemLoader(self.templates_folders)
        ])
        # env = Environment(loader=file_loader)
        env = Environment(loader=loader,
                          autoescape=select_autoescape())
        env.trim_blocks = True
        env.lstrip_blocks = True
        env.rstrip_blocks = True
        globals_obj = request.context.get_objs()
        for k, v in globals_obj.items():
            env.globals[k] = v
        template = env.get_template(content)
        # print(**request.context.get_objs())
        content = template.render(
            **{'user': request.user, 'messages': messages_broker.get_messages(request), **request.params, **context})
        # else:
        #     content = f'Template {content} not found.'
        # content = template.render(*context)
        self.response_content.append(content.encode())

    # def is_available(self, filename):
    #     for path in self.templates_folders:
    #         if not path.endswith('/'):
    #             path = path + '/'
    #         if os.path.isfile(path + filename):
    #             return filename
        # return None


class HttpResponse(Response):

    def __init__(self, request: Request, content,  status_code='200 OK', content_type='text/html'):
        super().__init__(request, status_code, content_type)
        if type(content) == str:
            content = f'{content}'.encode()
        self.response_content.append(content)


class JsonResponse(Response):

    def __init__(self, request: Request, content: str,  status_code='200 OK', content_type='application/json'):
        super().__init__(request, status_code, content_type)
        import json
        content = json.dumps(content).encode()
        self.response_content.append(content)


class ErrorResponse(Response):
    def __init__(self, request: Request, error_code: str):
        super().__init__(request, '404 Not Found', 'text/html')
        # self.response_content.append('Server Error'.encode())


class Http404Response(ErrorResponse):
    def __init__(self, request: Request, debug, routes):
        super().__init__(request, '404 Not Found')
        logs = []
        for p in routes:
            r = p.path
            if p.name:
                r += ' ' + p.name
            logs.append(r)
        doc = ErrorPage(request, logs, '404 page not found',
                        'available paths', '', debug)
        self.response_content.append(str(doc).encode())


class Http500Response(ErrorResponse):
    def __init__(self, request: Request, message, debug):
        super().__init__(request, '500 Server Error')
        doc = ErrorPage(request, [], '500 Server Error', message, '', debug)
        self.response_content.append(str(doc).encode())
        # self.response_content.append(f'500 Server Error \n {message}'.encode())


class Redirect(Response):
    def __init__(self, request: Request, route,  status_code='302 Found', content_type='text/html', *args, **kwargs):
        super().__init__(request, status_code, content_type)
        self.headers += [('Location', tesla.functions.url(route, **kwargs))]
        # print(self.headers)



class JavaScript:
    def document(self):
        pass

class Document:
    def get(self, target):
        return Event(f"document.querySelector('{target}')")
    
    def alert(self, obj):
        return f"alert('{obj}')"
    
    def log(self, obj):
        return f"console.log('{obj}')"

    

document = Document()


class Event:
    def __init__(self, target):
        self.target = target
    def addEventListener(self, event, func = None):
        return f"{self.target}.addEventListener('{event}', ()=>{func});" 
        
    def innerElem(self, elem):
        return f"{self.target}.innerHTML = '{elem}'" 

    def onClick(self, event):
        return self.addEventListener('click', event)
    
    def remove(self, target):
        return f"{self.target}.removeChild({self.target}.querySelector('{target}'))"

    def append(self, child):
        return f"{self.target}.append('{child}')"
    
    



from tesla.router import Path
from tesla.response import HttpResponse
from dataclasses import dataclass
import os


media_types = {
    'video': ['mp4'],
    'audio': ['mp3'],
    'image': ['png', 'jpg', 'jpeg'],
    'text': ['css', 'js', 'txt','json']
    
}

def get_media_type(ex):
    r = 'text/txt'
    
    for t, es in media_types.items():
        
        if ex.lower() in es:
            r = f'{t}/{ex}'
            
            break
    return r    
@dataclass
class StaticFiles:
    paths = ['./static/']
    request = None
    url = '/static/'



    def __post_init__(self):
        self.urls =  [
                Path('/{filename}/', self.file)     
            ]


    def file(self, request):
        # print(request.params)
         
        filename = request.params['filename']
        # print(self.paths)
        for path in self.paths:
            # print(path + filename)
            if os.path.isfile(path + filename):
                # print(path)
                filename = path + filename
                with open(filename, 'rb') as file:
                    # print(file.)
                    # request.headers += [('cache-control', 'public, max-age=3256926')]
                    return HttpResponse(request, file.read(), content_type=get_media_type(filename.split(".")[-1]))
            
        return HttpResponse(request, f"invalidd address\n {filename}")


 
staticfiles = StaticFiles()



from tesla.response import Redirect, tesla
from tesla.static import staticfiles
from tesla.router import router

from tesla.pyhtml import CT
# from

# path = f'http://{staticfiles.request.http_host}'


def media_file(url, *args, **kwargs):
    # print(url)
    if url == None:
        return ''
    _, filetype = url.split('>')
    f, _ = filetype.split('/')
    # url =  url
    if f == 'video':
            return CT('video',*args, **kwargs,  src=url)
    elif f == 'audio':
            return CT('audio',*args, **kwargs, src=url)
    elif f == 'image':
            return CT('img',*args, **kwargs, src=url)
            
            

def truncate(sentence, n=15):
    if len(sentence) > n:
        return sentence[:n]+'...'
    return sentence


def include():
    ...

def static_url(file_path):
   
    complete_url = staticfiles.url + file_path
    return complete_url    


def url(addr, *args, **kwargs):
    for route in router.routes:
        if ':' in addr:
            # print(addr)
            *_, ap, ur = addr.split(':')
            addr = f'{ur}_{ap}'
        if route.name == addr:
            sp_u = route.path.split('/')
            for i,p in enumerate(sp_u):
                
                if "{" and "}" in p:
                    v = p[1:-1]
                    param = kwargs.get(v)
                    if not param:

                        raise Exception(f'''
    {v} can"t be None.
    it suppose to be url('{addr.replace('_', ':')}', {v}=*something* ...)

                            ''')
                    sp_u[i] = param
                    
            return f'http://{staticfiles.request.http_host}' + '/'.join(sp_u)
   
    return None



def redirect(request, path):
    return Redirect(request, path)



class BaseException(Exception):
    
    def __init__(self, *args: object) -> None:
        super().__init__(*args)
    ...
    

class FileException(BaseException, FileNotFoundError):
    
    def __init__(self, *args: object) -> None:
        super().__init__(*args)
        
class TemplateException(FileException):
    
    def __init__(self, *args: object) -> None:
        super().__init__(*args)            

class MediaException(FileException):
    
    def __init__(self, *args: object) -> None:
        super().__init__(*args)    

class DatabaseException(BaseException):
    
    def __init__(self, *args: object) -> None:
        super().__init__(*args)
        

class KeyValueErrorException(DatabaseException, ValueError):
    
    def __init__(self, *args: object) -> None:
        super().__init__(*args)            
        
        
        

# raise KeyValueErrorException('hello')        


from dataclasses import dataclass, field
from datetime import datetime
import json
import os



@dataclass
class Session:
    values : dict = field(default_factory=dict)
    
    
    def __post_init__(self):
        if os.path.isfile('session.db'):
            with open('session.db') as file:
                obj = json.load(file)
                self.values = {**self.values, **obj}

    def add_to_session(self, key, value):
        today = datetime.today()
        expires = datetime(today.year, today.month, today.day).timestamp() + 5000000
        for k, v in self.values.items():
            if v == value:
                del self.values[k]
                break
        self.values[key] = {**value, 'expire_at':expires}
        self.save()

    def get_values(self):
        return self.values.values()
    
    def get(self, key):
        print(f'{key=}')
        if key not in self.values:
            return None
        if self.values.get(key)['expire_at'] <= datetime.today().timestamp():
            delattr(self.values, key)
            return None
        return self.values.get(key)
    
    
    def save(self):
        with open('session.db', 'w') as file:
            json.dump(self.values, file, indent=4)


        


from dataclasses import dataclass, field


@dataclass    
class Message:
    seen : bool = False
    message : str = ''
    level : int = 1
      
    def __str__(self) -> str:
        self.seen = True
        return self.message
 
@dataclass    
class MessageBroker:
    messages : list = field(default_factory=list) 
    
    def add_message(self,req, message, level=1):
        
        m = Message(message=message, level=level)
        b = self.get_messages(req)
        b.add_message(m)
        
    def get_messages(self, req):
        for m in self.messages:
            if m._req_s == req.session_id:
                return m
        m = Messages()
        m._req_s = req.session_id  
        self.messages.append(m)
         
        return m        
    
    
        
        
messages_broker = MessageBroker()
        

@dataclass
class Messages:
    __messages : list = field(default_factory=list)      
    _req_s : str = None  
    def add_message(self,m):

        self.__messages.append(m)
    def remove_seen(self):
        self.__messages = list(filter(lambda x: x.seen == False, self.__messages))     
    def __iter__(self):
        self.remove_seen()
        self.current_index = 0
          
        return self
    
    def __next__(self):
        if self.current_index < len(self.__messages):
            # ls = [x for x in self.messages if x.req_s == self.req_s]
            self.remove_seen()
            x = self.__messages[self.current_index]
            self.current_index += 1
            return x
        else:
            raise StopIteration
        
        
        
          



def dot(n):
    return " "*n
def lastIndex(txt, vl):
    return txt.index(vl)
def format(text):
   # text = (text.replace(" ",""))
    text = text.strip()
    lst = []
    indent = 0
    while len(text) > 0:
        index = lastIndex(text, ">")
        if text.startswith("<"):
            lst.append(text[:index+1])
            text = text[index+1:]  
        else:
          
             ind = lastIndex(text, "<")
             lst.append(text[:ind].strip())
             text = text[ind:]
    prev = ""
    #while lst.index(""):
       # (lst.remove(""))
    #print(dir(lst))
    
    for i in lst:
         if i == "":
             continue
         if i.startswith("</") and prev.startswith("</"):
             indent -= 1   
         elif prev.startswith("<") == False and i.startswith("</") == False:
             pass
         elif prev.startswith("</") and i.startswith("<"):
             pass
         elif prev.startswith("</") and i.startswith("<") == False:
             pass
         elif prev.startswith(("<meta","<img","<input","<link")) and i.startswith("</head"):
             indent -= 1
         elif prev.startswith(("<meta","<img","<input","<link")):
             pass

         elif prev.startswith("<") == False and i.startswith("</"):
             indent -= 1
   
         else:
             indent += 1
         text += f"{dot(indent)}{i}\n"
         prev = i
         
    
    return text
    


import random as r
import string

from tesla import TeslaApp
from tesla.auth.forms import LoginForm, RegisterForm
from tesla.database.localdb import global_db
from tesla.functions import redirect
from tesla.pyhtml import PYHTML
from tesla.pyhtml.tags import CSS, CT, CSSGroup
from tesla.response import HttpResponse


def login(request, user):
    session = ''.join(r.sample(string.ascii_letters, 50))

    request.set_cookie('user_session', session, {'max-age': 1296000})
    global_db.add(session, user)

    request.session.add_to_session(
        session, {'__id': user.id, 'session': str(session)})


def logout(request):
    request.headers += [('Set-Cookie',
                        'user_session=null;expires=Thu,01 Jan 1970 00:00:00 GMT;path=/')]


def LoginView(request):

    if request.is_authenticated:
        return redirect(request, '/')
    form_data = LoginForm()

    if request.method == 'POST':
        username = request.post.get('username')
        password = request.post.get('password')
        user = TeslaApp.auth_model.get(username=username, password=password)
        # print(user)
        if user:
            login(request, user)
            return redirect(request, '/')
        # print(username, password)
    doc = PYHTML()
    head, body = doc.create_doc()

    # head
    title = CT('title', 'Tesla | Authentication View')
    style = CT('style', '''
               *{
    box-sizing: border-box;
    margin: 0;
}

form{
    width: 320px;
    height: fit-content;
    border: 1px solid;
    display: flex;
    flex-direction: column;
    gap: 10px;
    padding: 15px;
    font-size: large;
    margin: 20px auto;
    
    background-color: #2196F3;
    color: white;
}
input{
    width: 100%;
    padding: 10px;
    border: 1px solid lightgray ;
    border-radius: 5px;
    outline: none;
    
    background-color: #fafafa;
}
#rmb{
    display: flex;
    align-items: center;
    width: 100%;
    gap: 10px;
}
#rmb > input{
    width: fit-content;
}
button{
    padding: 10px;
    border-radius: 5px;
    border: none;
    background-color: #26ff00;
    cursor: pointer;
}

               ''')
    p = CT('p', '*This is default Tesla login page')
    form_style = CSS(
        display='flex',
        flexDirection='column',
        alignItems='center',
        padding='10px',
        margin='auto',
        marginTop='30px',
        width='250px',
        gap='10px',
        backgroundColor='lightgray',
        borderRadius='5px'
    )

    head.append(title, style)
    # form
    form = CT('form', method='POST')
    csrf = CT('input', value=request.csrf, name='csrfmiddleware', hidden=1)
    h2 = CT('h2', 'User login view')

    login_btn = CT('button', 'login')
    form.append(h2, csrf, form_data, login_btn, p)
    body.append(form)
    return HttpResponse(request, str(doc))


def RegisterView(request):

    if request.is_authenticated:
        return redirect(request, '/')
    form_data = RegisterForm()

    if request.method == 'POST':
        username = request.post.get('username')
        password = request.post.get('password')
        user = form_data.validate(request.post, TeslaApp.auth_model)
        # print(request.post.get('rmb'))
        # user = User(username=username, password=password).save()
        # print(user)
        if user:
            login(request, user)
            return redirect(request, '/')
        # print(username, password)
    doc = PYHTML()
    head, body = doc.create_doc()

    # head
    title = CT('title', 'Tesla | Authentication View')
    style = CT('style', '''
               *{
    box-sizing: border-box;
    margin: 0;
}

form{
    width: 320px;
    height: fit-content;
    border: 1px solid;
    display: flex;
    flex-direction: column;
    gap: 10px;
    padding: 15px;
    font-size: large;
    margin: 20px auto;
    
    background-color: #2196F3;
    color: white;
}
input{
    width: 100%;
    padding: 10px;
    border: 1px solid lightgray ;
    border-radius: 5px;
    outline: none;
    
    background-color: #fafafa;
}
#rmb{
    display: flex;
    align-items: center;
    width: 100%;
    gap: 10px;
}
#rmb > input{
    width: fit-content;
}
button{
    padding: 10px;
    border-radius: 5px;
    border: none;
    background-color: #26ff00;
    cursor: pointer;
}

               ''')
    p = CT('p', '*This is default Tesla Register page')

    head.append(title, style)
    # form
    form = CT('form', method='POST', enctype="multipart/form-data")
    csrf = CT('input', value=request.csrf, name='csrfmiddleware', hidden=1)
    h2 = CT('h2', 'User Register view')

    login_btn = CT('button', 'login')
    form.append(h2, csrf, form_data, login_btn, p)
    body.append(form)
    return HttpResponse(request, str(doc))


def authenticate(user, en_password):
    from argon2 import PasswordHasher
    ph = PasswordHasher()
    try:
        ph.verify(user.password, en_password)
        if ph.check_needs_rehash(user.password):
            user.password = ph.hash(password)
            user.save()

            ...
        return True
    except:
        return False



from dataclasses import dataclass, field, asdict, astuple
# from .auth import csrf_middleware

@dataclass
class Middlewares:
    _middlewares : list = field(default_factory=list, init=False, repr=False)
    
    
    def set_middlewares(self, middlewares):
        for middleware in middlewares:
            self.register_middleware(middleware) 
            
    def register_middleware(self, middleware):
        if middleware not in self._middlewares:
            self._middlewares.append(middleware) 
            
            
    def middlewares(self):
        return self._middlewares
    

    
    
    
middlewares = Middlewares()    

# middlewares.register_middleware(csrf_middleware)


from dataclasses import dataclass
from tesla.pyhtml.tags import CT
from tesla.modal import ForeignKey, ManyToManyField
from tesla.request import TemporaryFile
import copy


class Form:
    def validate(self, obj, model=None):
        data = dict(obj)
        del data['csrfmiddleware']
        for key, value in data.items():
            # print(key, value)
            b = self.fields()[key].validate(value)
            if not b:
                ...
        if model:
            return model(**data).save()        
        return True    
             
        # print(self.fields())
    def save(self):
        fields = self.fields()
        # print(self.html())
    
    def fields(self):

        return self.__dict__

    def html(self):
        return ''.join(map(lambda f: f.field(), self.fields().values()))

    def __repr__(self):
        return self.html()

    def __str__(self):
        return (self.html())
        
class ModelForm:
    
    def __init__(self, instance = None) -> None:
        self.instance = instance
        
    
    def html(self):
        html = []
        model = self.model()
        if isinstance(self.fields, list):
            fields = self.fields
        else:
            fields = self.model.property_cls()   
        for field in fields:
            value = copy.deepcopy(getattr(model, field))
            value.name = field
            
            if self.instance:
                obj = self.instance
                # print(value.default)
                default = (getattr(obj, field))
                if type(default) in [int, bool, float, str]:
                    value.default = default
            else:
                if issubclass(type(value), (ForeignKey)):
                    
                        value.related_model_id = None
                    
                if issubclass(type(value), (ManyToManyField)):
                    value.related_model_ids = []            
            #     value.default = ''        
            html.append(value)
            ...
        # print('passs......')    
        return html    
            
    def __str__(self) -> str:
        return f'{self.html()}'   

    def validate(self, **kwargs):
        
        new_model = copy.deepcopy(self.model)

        if type(self.fields) == list:
            if self.fields != list(kwargs.keys()):
                not_kwargs = ', '.join([a for a in set(self.fields) if a not in set(kwargs.keys())])
                raise Exception(f'{not_kwargs} are not valid fields for {self.model.__name__}')
            for field in self.fields:
                f = getattr(new_model, field)
                v = f.validate(kwargs.get(field))
                
            return new_model.create(**kwargs) 

        for field, value in kwargs.items():
            if '__' in field:
                field = field.split('__')[0]
                tt = type(getattr(new_model, field))
                if not issubclass(tt, (ForeignKey, ManyToManyField)):
                    field += '__' + field[1]
                    
            f = getattr(new_model, field)
            v = f.validate(value)
        # print(kwargs)  
        # print('123')  
        return new_model.create(**kwargs) 

    def save(self):


        ...             
    
    

 



@dataclass
class Field:
    placeholder : str = 'Input'
    type : str = 'text'
    data_type : str = 'text'
    name : str = 'name'
    label: bool = True
    label_text : str = ''
    label_id : str = ''
    

    def __str__(self):
        return  self.field()

    def __repr__(self):
        return self.__str__()
    
    def field(self):
        input = CT('input', type=self.type, placeholder = self.placeholder, name=self.name)
        if self.label:
            if self.label_text == '':
                self.label_text = self.name
            input = CT('label', self.label_text , input, id=self.label_id)
        return (input.html())
    
    def validate(self, value):
        return isinstance(value, self.data_type)




from typing import Any, Dict, List, AnyStr, Set

class __Localdb:
    
    def __init__(self) -> None:
        self.__db = {}
        
    # def save(self):
    def get(self, key : AnyStr, ps=True) -> Any:
        if key in self.__db or ps:
            return self.__db.get(key)
        raise KeyError(f'{key} not found.') 
    
    def clear(self, key : AnyStr):
        if key in self.__db:
            delattr(self.__db, key)
            return None
        raise KeyError(f'{key} not found.') 
    
    def add(self, key : AnyStr, value : Any, ps=True):
        if key in self.__db and not ps:
            raise Exception(f'{key} already exists.')
        self.__db[key] = value
        return self.get(key)
    
    def exists(self, key : AnyStr) -> bool:
        if key in self.__db:
            return True
        return False
    
    

global_db = __Localdb()    




class Search:
  
    
    def __init__(self, model = None, search_for = None, lookups = None, exact=False):
        self.model = model
        self.search_for = search_for.strip()
        self.lookp_ups = lookups
        self.exact = exact
        if self.model:
            self.objs = self.model
    
    def result(self):
        if len(self.objs) == 0:
            return []
        temp = set()
        for model in self.objs:
            # print(model)
            for f in self.lookp_ups:
                
                if not hasattr(model, f):
                    continue
                
                if not self.exact:
                    if self.search_for.lower() in getattr(model, f).lower():
                        temp.add(model)
                        break
                else:
                    if getattr(model, f).lower() == self.search_for.lower():
                        temp.add(model)
                        break
                    
        return temp            



import tesla
from tesla.auth.decorators import login_required
from tesla.auth.views import login as user_login, logout as user_logout, authenticate
from tesla.functions import url
from tesla.response import HttpResponse, Redirect, Render, JsonResponse
from tesla.pyhtml import CT
from .models import User
from .forms import AdminForm
from tesla.database.jsondb import JsonDB
# from tesla.views import View, DetailView, RetriveAllView
from tesla.pagination import Paginator
from tesla.search import Search

from . import collections_manage

# your views


@login_required(path='admin:login')
def collections(request):
    context = {}
    context['names'] = collections_manage.colls()

    return Render(request, 'admin/colls.html', context)


@login_required(path='admin:login')
def collection(request):
    collection = request.params.get('collection')
    model_db = JsonDB(collection + '/')
    model = model_db.model()
    context = {}
    search_for = request.query.get('search')
    if search_for:
        models = Search(model.all(models=True), search_for,
                        model.__meta__()).result()
        context['search_for'] = search_for
    else:
        models = model_db.all(models=True)
    page = request.query.get('page')
    if not page:
        page = 1
    paginator = Paginator(models, page=int(page))
    # print(models)
    context['objs'] = paginator.current()
    context['next'] = paginator.next()
    context['previous'] = paginator.previous()
    context['pages'] = paginator.pages_lists()
    context['page'] = paginator.page
    # print(paginator.pages_lists())
    # context['objs'] = models
    context['info'] = model.__meta__()
    context['collection'] = collection
    return Render(request, 'admin/col.html', context)


@login_required(path='admin:login')
def collection_obj(request):
    # print(request.user.is_super_user)
    lookup = request.params.get('lookup')
    collection = request.params.get('collection')
    model_db = JsonDB(collection + '/')
    obj = model_db.model().get(id=lookup)
    if 'json' in request.query:
        return JsonResponse(request, obj.json())

    form = AdminForm(obj)
    form.model = model_db.model
    form.fields = '__all__'
    if request.method == 'POST':
        data = request.post.data
        del data['csrfmiddleware']
        # print(data)
        obj.update(**data)
        obj = obj.save()

        return Redirect(request, 'admin:collection_obj', lookup=lookup, collection=collection)
    context = {}
    context['lookup'] = lookup
    context['collection'] = collection
    context['form'] = form
    return Render(request, 'admin/obj.html', context)


@login_required(path='admin:login')
def collection_new(request):
    context = {}
    collection = request.params.get('collection')
    model_db = JsonDB(collection + '/')
    model = model_db.model()
    form = AdminForm()
    form.model = model_db.model
    form.fields = '__all__'
    context['form'] = form
    if request.method == 'POST':
        data = request.post.data
        del data['csrfmiddleware']
        f = form.validate(**data)
        f.save()
        return Redirect(request, 'admin:collection_obj', collection=collection, lookup=f.id)
    return Render(request, 'admin/col_n.html', context)


@login_required(path='admin:login')
def collection_del(request):
    lookup = request.params.get('lookup')
    collection = request.params.get('collection')
    model_db = JsonDB(collection + '/')
    obj = model_db.model().get(id=lookup)
    obj.delete()
    return Redirect(request, 'admin:collection', collection=collection)


@login_required(path='admin:login')
def collection_del_all(request):
    lookup = request.params.get('lookup')
    collection = request.params.get('collection')
    model_db = JsonDB(collection + '/')
    objs = model_db.model().all(models=True)
    for obj in objs:
        obj.delete()
    return Redirect(request, 'admin:collection', collection=collection)


@login_required(path='admin:login')
def collection_download(request):

    collection = request.params.get('collection')
    model_db = JsonDB(collection + '/')
    objs = model_db.model().all(models=False)
    return JsonResponse(request, objs)


@login_required(path='admin:login')
def admin_account(request):
    obj = tesla.TeslaApp.auth_model.get(id=request.user.id)
    if 'json' in request.query:
        return JsonResponse(request, obj.json())
    # print(obj)
    form = AdminForm(obj)
    form.model = tesla.TeslaApp.auth_model
    form.fields = '__all__'
    if request.method == 'POST':
        data = request.post.data
        del data['csrfmiddleware']
        obj.update(**data)
        obj.save()
    context = {}
    context['form'] = form
    return Render(request, 'admin/account.html', context)


@login_required(path='admin:login')
def admin_settings(request):
    ...


@login_required(path='admin:login')
def index(request):
    context = {}
    context['names'] = collections_manage.colls()
    # print(context)
    return Render(request, 'admin/base_admin.html', context)


def login(request):

    if request.method == 'POST':
        u = request.post.get('username')
        p = request.post.get('password')
        user = User.get(username=u)
        if authenticate(user, p):
            user_login(request, user)
            return Redirect(request, 'admin:index')
    return Render(request, 'admin/login.html')


def register(request):
    if request.method == 'POST':
        u = request.post.get('username')
        e = request.post.get('email')
        p = request.post.get('password')
        User.create(username=u, email=e, password=p)
        return Redirect(request, 'admin:login')
    return Render(request, 'admin/register.html')


def reset_password(request):
    if request.method == 'POST':
        # print('es')
        pass
    return Render(request, 'admin/reset-password.html')


def logout(request):
    user_logout(request)
    return Redirect(request, 'admin:login')




import os

import time
class Watcher:
    def __init__(self, dir, generate_files, server):
        self.dir = dir
        self.files = os.listdir(dir)
        self.generate_files = generate_files
        self.server = server
        self.start_watching()
    def start_watching(self):
        files = {}
        while True:
            
            
            for i, file in enumerate(self.files):
               
                if os.path.isdir(file):
                    continue
                with open(self.dir + file, "r") as f1:
                     if files.get(file):
                        f2 = files.get(file)
                     else:
                       
                        files[file] = f1.read()
                        f2 = f1.read()
                     
                     
                     if f1.read() != f2:
                            files[file] = f1.read()
                            print(f"{file} changed")
                            print('reloading the server...')
                            self.generate_files()
                            self.server.server_close()
                            self.server.shutdown()
                            os.system( 'python main.py' )
                            
                            


from dataclasses import dataclass
from tesla.form import Field, Form


@dataclass
class LoginForm(Form):
    username : str = Field(placeholder = 'Enter username', name='username', label_text='Username',data_type = str)
    password : str = Field(placeholder = 'Enter password', name= 'password', type = 'password', label_text='Password', data_type = str)
    remember_me : bool = Field(placeholder='Remember Me', type='checkbox', name='rmb', label_text='Remember Me', label_id='rmb', data_type = bool)

@dataclass
class RegisterForm(Form):
    username : str = Field(placeholder = 'Enter username', name='username', label_text='Username', data_type = str)
    password : str = Field(placeholder = 'Enter password', name= 'password', type = 'password', label_text='Password', data_type = str)
    confirm_password : str = Field(placeholder = 'Enter password', name= 'confirm_password', type = 'password', label_text='Confirm Password', data_type = str)
    remember_me : bool = Field(placeholder='Remember Me', type='checkbox', name='remember_me', label_text='Remember Me', label_id='rmb', data_type = bool)






class Signal:
    __signals = {}
    
    
    def connect(self, sender, receiver, signal='pre'):
        sign = sender.__name__ + signal
        rs = self.__signals.get(sign)
        if rs:
            self.__signals[sign] += [receiver]
            return
        self.__signals[sign] = [receiver]
    
    def send(self, sender, instance, created,signal, **kwargs):
        sign = sender.__name__ + signal
        rs = self.__signals.get(sign, [])
        for r in rs:
            r(sender, instance, created,**kwargs)         
            
signal = Signal()

def connect_pre_save(sender, receiver):
    signal.connect(sender=sender, receiver=receiver, signal='pre-save')

def connect_post_save(sender, receiver):
    signal.connect(sender=sender, receiver=receiver, signal='post-save')    


from dataclasses import dataclass, field
from typing import Any
import os
import json
# from uuid import uuid4 as v4
import copy
# sorted
from tesla.database.jsondb.schema import User, settype
import tesla

ABS_PATH = './db/'


def Contains(column, obj):
    r = {}
    # print(obj)
    for k in obj.keys():
        r[k] = False
        if column[k] == obj[k]:
            r[k] = True
    for c in list(r.values()):
        if c == False:
            return False
    return True


def GreaterThan(columns, obj, k, op):
    key = k.find(f"__{op}t")
    # print(k)
   # print(columns["age"] > obj[k], columns)
    if op == 'g':
        statement = columns[k[:key]] > obj[k]
    else:
        statement = columns[k[:key]] < obj[k]
    # print(statement)
    if statement:

        o = {}
        for ky, v in obj.items():

            if ky.find(f"__{op}t") != -1:
                o[ky[:key]] = columns[k[:key]]
            else:
                o[ky] = v

            # print(o)
        return columns, o, True
        pass
    return columns, obj, False


@dataclass
class ColCache:
    obj: Any
    db: dict = field(default_factory=dict)

    def update(self, collection, key, value):
        if self.db.get(collection):
            self.db[collection][key] = value
        raise Exception(f'No collection with name {collection}')

    def get(self, collection, key=None):
        if key:
            return self.db.get(collection).get(key)
        return self.db.get(collection)

    def set(self, key, json):
        if self.get(key) == None:
            self.db[key] = dict()
        self.db[key] = (json)

    def delete(self, key):
        del self.db[key]

        # print(key)
        # print(self.db)
local_c_db = ColCache({})


@dataclass
class DBCache:
    db: dict = field(default_factory=dict)

    def update(self, collection, key, value):
        if self.db.get(collection):
            self.db[collection][key] = value
        else:
            self.db[collection] = dict()
            self.db[collection][key] = value

    def get(self, collection, key=None):
        if key:
            o = self.db.get(collection)
            if o:
                j = o.get(key)
                if j:
                    return j.values()
                return o.values()
        o = self.db.get(collection)
        return o.values() if o else []

    def remove(self, collection, key):
        del self.db[collection][key]


local_db = DBCache({})


# @dataclass
class Column:

    def __init__(self, file):
        self.file = file
        self.__one_2_many = {}
        # self.many_2_many = []

        *_, self.cls, id = file.split('/')
        self.id = id.split('.')[0]

    def __post_init__(self):
        for k, v in self.readAll().items():
            # print(k, v)
            if '__' in k:
                m, k = k.split('__')
                self.__one_2_many[m] = {k: v}
        ...

    def update(self, **obj):
        obj = dict(obj)
        json_db = local_c_db.get(self.id)
        if not json_db:
            with open(self.file) as f:
                json_db = json.load(f)
        for key, value in obj.items():
            json_db[key] = value
        with open(self.file, 'w+') as f:
            json.dump(json_db, f, indent=4)
            local_c_db.set(self.id, json_db)

    def readAll(self, dept=1):
        json_db = {}
        if local_c_db.get(self.id):
            json_db = local_c_db.get(self.id)

        else:
            with open(self.file) as f:
                json_db = json.load(f)

                local_c_db.set(self.id, json_db)
        if dept > 1:
            # print(dept)
            self.__post_init__()
            for key in self.__one_2_many:
                obj = self.__one_2_many[key]
                json_db[key] = self.get_foreign_obj(key, **obj).readAll()
        return json_db

    def get(self, key):
        # print(key)
        # print(self.one_2_many)
        if self.__one_2_many == {}:
            self.__post_init__()

        if key in self.__one_2_many:
            obj = self.__one_2_many[key]
            return self.get_foreign_obj(key+'/', **obj)
        if local_c_db.get(self.id):
            return local_c_db.get(self.id, key)

        with open(self.file) as f:
            json_db = json.load(f)
            local_c_db.set(self.id, json_db)
        return json_db.get(key)

    def clear(self, key):
        if local_c_db.get(self.id):
            local_c_db.delete(self.id)
        with open(self.file, 'r+') as file:
            json_db = json.load(file)
            if type(key) == list:
                for k in key:
                    del (json_db[k])
            else:
                del (json_db[key])
        return None

    def delete(self):
        local_db.remove(self.cls, self.id)
        if local_c_db.get(self.id):
            local_c_db.delete(self.id)
        os.remove(self.file)

    def clearAll(self):
        if local_c_db.get(self.id):
            local_c_db.set(self.id, {})
        with open(self.file, 'w+') as file:
            json.dump({}, file, indent=4)
        return None

    def __repr__(self):
        return f"Column {self.get('id')}"

    def get_foreign_obj(self, key, **args):
        return JsonDB(key.title()).get(args)


class DB:
    # db = {}
    abs_path = ''

    def __init__(self, collection, path='./db/'):
        self.collection = collection
        self.abs_path = path
        self.path = path + 'collections/'
        self.model = None
        ABS_PATH = self.abs_path
        self.__metas = None

        if os.path.isdir(self.path + self.collection) == False:
            os.makedirs(self.path + self.collection)

        self.__model__()

    def __model__(self):
        model_n = self.collection[:-1]
        for m in tesla.TeslaApp.registered_models:

            if m.__name__ == model_n:
                self.model = m

    def size(self):
        l = os.listdir(self.path + self.collection)
        return len(l), sum(os.path.getsize(self.path + self.collection + f) for f in l if os.path.isfile(self.path + self.collection + f))

    @classmethod
    def get_collection(cls, model, dept=1):

        if model.title() not in cls.db()['collections']:
            return {
                'Error': f'No such collection {model}'
            }
        # return type(model, (), )
        return JsonDB(model.title() + '/').all()

    @classmethod
    def db(cls):
        js = {}
        return json.load(open(ABS_PATH + 'db.json'))

    def create_column(self, model, table_name):
        bck = None
        if os.path.isfile(self.path + self.collection + table_name + '.json'):
            with open(self.path + self.collection + table_name + '.json') as file:
                bck = json.load(file)
        try:
            
            with open(self.path + self.collection + table_name + '.json', 'w+') as file:
                json.dump(model, file, indent=4)
                local_db.update(self.collection.split(
                    '/')[0], model.get('id'), model)
                return None
        except Exception as e:
            if bck:
                with open(self.path + self.collection + table_name + '.json', 'w+') as file:
                    json.dump(bck, file, indent=4)
            raise Exception(e)

    def delete(self, table_name):
        os.remove(self.path + self.collection + table_name + '.json')
        local_db.remove(self.collection[:-1], table_name)
        return None

    def all(self, models=False):
        # print(local_db)

        lookup = self.collection.split('/')[0]
        if local_db.get(lookup):
            if models:
                ls = []
                for js in local_db.get(lookup):

                    c = self.model()
                    for k, v in js.items():
                        setattr(c, k, v)
                    ls.append(c)

                return ls
            # print(local_db.get(lookup))

            return local_db.get(lookup)

        all = os.listdir(self.path + self.collection)
        for i in all:

            with open(self.path + self.collection + i) as fl:
                js_ = json.load(fl)

                local_db.update(lookup, js_.get('id'), js_)
        if models:
            ls = []
            for js in local_db.get(lookup):
                c = self.model()
                for k, v in js.items():
                    setattr(c, k, v)
                ls.append(c)
            return ls
        # print(local_db.get(lookup))
        return local_db.get(lookup)

    def get(self, obj):
        all = self.all()
        user = None
        for i in all:
            a = 0
            b = len(obj)

            for k, v in obj.items():

                if i.get(k) == v:
                    a += 1
            if a == b:
                user = i
                break
        return user

    def filter(self, ob):
        all = self.all()
        result = []

        for i in all:
            obj = copy.deepcopy(ob)
            o = i.readAll()
            # print(obj)
            keys = set(o.keys())
            keys2 = set(obj.keys())
            gt = None
            for k in keys2:
                if k.find("__") != -1:
                    op = 'g' if k.find("__gt") != -1 else 'l'
                    # print(op)
                    o, obj, gt = GreaterThan(o, obj, k, op)
                    keys2 = set(obj.keys())
                    # print(obj)
            if gt == False:
                # print(gt)
                continue
            t = (keys2 & keys)

            if list(t) == list(keys2):
                contains = Contains(o, obj)
                # print(contains)
                if contains:
                    result.append(i)
        return result
